import unittest

from artap.problem import Problem
from artap.algorithm_scipy import ScipyOpt
from artap.algorithm import EvaluatorType


class RobustnessProblem(Problem):
    """ Describe simple one objective optimization problem. """

    def set(self):
        self.name = "ComsolProblem"
        self.parameters = [{'name': 'a', 'initial_value': 10, 'tol': 1e-1},
                           {'name': 'b', 'initial_value': 10, 'tol': 1e-1}]
        self.costs = [{'name': 'F1', 'criteria': 'minimize'}]

    def evaluate(self, individual):
        x_1 = individual.vector[0]
        x_2 = individual.vector[1]
        return [x_1**2 + x_2**2]


class TestSimpleOptimization(unittest.TestCase):
    """ Tests simple one objective optimization problem."""

    def test_gradient(self):
        problem = RobustnessProblem()
        algorithm = ScipyOpt(problem, evaluator_type=EvaluatorType.GRADIENT)
        algorithm.options['algorithm'] = 'Nelder-Mead'
        algorithm.options['tol'] = 1e-4
        algorithm.run()

        for individual in problem.populations[0].individuals:
            print(individual.features['gradient'])
            for child in individual.children:
                print(child)
            print("-------------------------")

    def test_worst_case(self):
        problem = RobustnessProblem()
        algorithm = ScipyOpt(problem, evaluator_type=EvaluatorType.WORST_CASE)
        algorithm.options['algorithm'] = 'Nelder-Mead'
        algorithm.options['tol'] = 1e-4
        algorithm.run()

        for individual in problem.populations[0].individuals:
            print(individual)
            for child in individual.children:
                print(child)
            print("-------------------------")


if __name__ == '__main__':
    unittest.main()



from .algorithm_genetic import GeneralEvolutionaryAlgorithm
import time


class SweepAlgorithm(GeneralEvolutionaryAlgorithm):
    """
    Sweep Analysis
    """

    def __init__(self, problem, generator, name='SweepAlgorithm'):
        super().__init__(problem, name)
        self.generator = generator

    def run(self):
        t_s = time.time()

        # create initial population and evaluate individuals
        population = self.gen_initial_population()
        self.evaluate(population.individuals)

        t = time.time() - t_s
        self.problem.logger.info("Sweep: elapsed time: {} s".format(t))



from artap.archive import Archive
from artap.individual import Individual

import unittest


class TestArchive(unittest.TestCase):

    def setUp(self):
        self.archive = Archive()

    def test_should_constructor_create_a_non_null_object(self):
        self.assertIsNotNone(self.archive)

    def test_should_adding_one_solution_work_properly(self):
        x = Individual([2, 2])
        x.costs = [-1.0, 5.0, 9.0]
        x.costs_signed = [-1.0, 5.0, 9.0, 0]

        self.archive.add(x)
        self.assertEqual(1, self.archive.size())
        self.assertEqual(x, self.archive._contents[0])

    def test_should_adding_two_solutions_work_properly_if_one_is_dominated(self):
        dominated_solution = Individual([1, 2])
        dominated_solution.costs = [2.0, 2.0, 0]
        dominated_solution.costs_signed = [2.0, 2.0, 0]

        dominant_solution = Individual([1, 1])
        dominant_solution.costs = [1.0, 1.0, 0]
        dominant_solution.costs_signed = [1.0, 1.0, 0]

        self.archive.add(dominated_solution)
        self.archive.add(dominant_solution)

        self.assertEqual(1, self.archive.size())
        self.assertEqual(dominant_solution, self.archive._contents[0])

    def test_should_adding_two_solutions_work_properly_if_both_are_non_dominated(self):
        x = Individual([1, 2])
        x.costs = [1.0, 0.0, 0]
        x.costs_signed = [1.0, 0.0, 0]

        y = Individual([1, 1])
        y.costs = [0.0, 1.0, 0]
        y.costs_signed = [0.0, 1.0, 0]

        self.archive.add(x)
        self.archive.add(y)

        self.assertEqual(2, self.archive.size())
        self.assertTrue(x in self.archive._contents and
                        y in self.archive._contents)

    def test_should_adding_four_solutions_work_properly_if_one_dominates_the_others(self):
        x = Individual([1, 2])
        x.costs = [1.0, 1.0]
        x.costs_signed = [1.0, 1.0, 0]

        y = Individual([1, 2])
        y.costs = [0.0, 2.0]
        y.costs_signed = [0.0, 2.0, 0]

        z = Individual([1, 2])
        z.costs = [0.5, 1.5]
        z.costs_signed = [0.5, 1.5, 0]

        v = Individual([1, 2])
        v.costs = [0.0, 0.0]
        v.costs_signed = [0.0, 0.0, 0]

        self.archive.add(x)
        self.archive.add(y)
        self.archive.add(z)
        self.archive.add(v)

        self.assertEqual(1, self.archive.size())
        self.assertEqual(v, self.archive._contents[0])

    def test_should_adding_three_solutions_work_properly_if_two_of_them_are_equal(self):
        x = Individual([1, 2])
        x.costs = [1.0, 1.0]
        x.costs_signed = [1.0, 1.0, 0.0]

        y = Individual([1, 2])
        y.costs = [0.0, 2.0]
        y.costs_signed = [0.0, 2.0, 0.0]

        z = Individual([1, 2])
        z.costs = [1.0, 1.0]
        z.costs_signed = [1.0, 1.0, 0.0]

        self.archive.add(x)
        self.archive.add(y)

        result = self.archive.add(z)

        self.assertEqual(2, self.archive.size())
        self.assertFalse(result)
        self.assertTrue(x in self.archive._contents
                        or y in self.archive._contents)



import unittest
import importlib.util
import os


def make_test_function(name, filename):
    def test(self):
        spec = importlib.util.spec_from_file_location(name, filename)
        script = importlib.util.module_from_spec(spec)
        path = filename.split(name + '.')[0]
        old_cwd = os.getcwd()
        os.chdir(path)
        spec.loader.exec_module(script)
        os.chdir(old_cwd)

    return test


class TestExamples(unittest.TestCase):
    """ Tests all examples in examples directory."""
    @staticmethod
    def init_tests():
        path = os.getcwd()
        for folder in os.walk(path):
            if folder[0].split('/')[-1] =='__pycache__':
                continue
            files = folder[2]
            for file in files:
                if file == '__init__.py' or file[-3:] != '.py':
                    continue
                else:
                    name = file.split('.')[0]
                    filename = folder[0] + os.sep + file
                    test_func = make_test_function(name, filename)
                    setattr(TestExamples, 'test_{0}'.format(name), test_func)


# create tests dynamically
TestExamples.init_tests()

if __name__ == '__main__':
    unittest.main()



from random import randint
from .problem import Problem
from .population import Population
from .algorithm_genetic import GeneralEvolutionaryAlgorithm
from .operators import SwarmMutator, DummySelector, RandomGenerator, SwarmMutatorTVIW

import time


class SwarmAlgorithm(GeneralEvolutionaryAlgorithm):

    def __init__(self, problem: Problem, name="General Swarm-based Algorithm"):
        super().__init__(problem, name)

    def iterate(self, parents, archive=None):
        """
        Calculates the speed and the new position of the individuals.
        :param parents:
        :param archive:
        :return:
        """
        offsprings = []
        offsprings.extend(parents)
        while len(offsprings) < 2 * self.population_size:
            parent1 = self.selector.select(parents)

            repeat = True
            while repeat:
                if archive:
                    parent2 = self.selector.select(archive)
                else:
                    parent2 = self.selector.select(parents)

                if parent1 is not parent2:
                    repeat = False

            # crossover
            child1, child2 = self.crossover.cross(parent1, parent2)

            # mutation
            child1 = self.mutator.mutate(child1)
            child2 = self.mutator.mutate(child2)

            if not any(child1 == item for item in offsprings):
                offsprings.append(deepcopy(child1))  # Always create new individual
            if not any(child1 == item for item in offsprings):
                offsprings.append(deepcopy(child2))  # Always create new individual

        # parents.extend(removed_parents)
        return offsprings

    def run(self):
        pass


class OMOPSO(SwarmAlgorithm):
    """
    Implementation of OMOPSO, a multi-objective particle swarm optimizer (MOPSO).
    OMOPSO uses Crowding distance, Mutation and ε-Dominance.
    According to [3], OMOPSO is one of the top-performing PSO algorithms.

    [1] Margarita Reyes SierraCarlos A. Coello Coello
        Improving PSO-Based Multi-objective Optimization Using Crowding, Mutation and ∈-Dominance
        DOI https://doi.org/10.1007/978-3-540-31880-4_35

    [2] S. Mostaghim ; J. Teich :
        Strategies for finding good local guides in multi-objective particle swarm optimization (MOPSO)
        DOI: 10.1109/SIS.2003.1202243
    [3] Durillo, J. J., J. Garc�a-Nieto, A. J. Nebro, C. A. Coello Coello, F. Luna, and E. Alba (2009).
        Multi-Objective Particle Swarm Optimizers: An Experimental Comparison.
        Evolutionary Multi-Criterion Optimization, pp. 495-509
    """

    def __init__(self, problem: Problem, name="Particle Swarm Algorithm"):
        super().__init__(problem, name)
        self.n = self.options['max_population_size']
        self.mutator = SwarmMutator(self.problem.parameters)
        self.selector = DummySelector(self.problem.parameters, self.problem.signs)
        self.features = {'velocity': [],
                         'best_position': [],
                         'max_speed': []}

        # set random generator
        self.generator = RandomGenerator(self.problem.parameters)

    def run(self):

        self.generator.init(self.options['max_population_size'])
        population = self.gen_initial_population()

        self.evaluate(population.individuals)
        self.add_features(population.individuals)

        for individual in population.individuals:
            self.mutator.evaluate_best_individual(individual)

        self.selector.fast_nondominated_sorting(population.individuals)

        t_s = time.time()
        self.problem.logger.info("PSO: {}/{}".format(self.options['max_population_number'],
                                                     self.options['max_population_size']))

        i = 0
        while i < self.options['max_population_number']:
            offsprings = self.selector.select(population.individuals)

            pareto_front = []
            for individual in offsprings:
                if individual.features['front_number'] == 1:
                    pareto_front.append(individual)

            for individual in offsprings:
                index = randint(0, len(pareto_front) - 1)  # takes random individual from Pareto front
                best_individual = pareto_front[index]
                if best_individual is not individual:
                    self.mutator.update(best_individual)
                    self.mutator.mutate(individual)

            population = Population(offsprings)
            self.problem.populations.append(population)
            self.evaluator.evaluate(offsprings)
            self.add_features(offsprings)

            for individual in offsprings:
                self.mutator.evaluate_best_individual(individual)

            self.selector.fast_nondominated_sorting(offsprings)

            i += 1

        t = time.time() - t_s
        self.problem.logger.info("PSO: elapsed time: {} s".format(t))


class PSO_V1(SwarmAlgorithm):
    """

    X. Li. A Non-dominated Sorting Particle Swarm Optimizer for Multiobjective
    Optimization. In Genetic and Evolutionary Computation - GECCO 2003, volume
    2723 of LNCS, pages 37–48, 2003.

    This algorithm is a variant of the original PSO, published by Eberhart(2000), the origin of this modified algorithm,
    which constriction factor was introduced by Clercs in 1999.

    The code is based on SHI and EBERHARTS algorithm.

    Empirical study of particle swarm optimization,” in Proc. IEEE Int. Congr. Evolutionary Computation, vol. 3,
    1999, pp. 101–106.
    """

    def __init__(self, problem: Problem, name="Particle Swarm Algorithm - with time varieting inertia weight"):
        super().__init__(problem, name)
        self.n = self.options['max_population_size']
        self.mutator = SwarmMutatorTVIW(self.problem.parameters, self.options['max_population_number'])
        self.selector = DummySelector(self.problem.parameters, self.problem.signs)

    def run(self):
        # set random generator
        self.generator = RandomGenerator(self.problem.parameters)
        self.generator.init(self.options['max_population_size'])

        population = self.gen_initial_population()
        self.evaluate(population.individuals)
        self.add_features(population.individuals)

        for individual in population.individuals:
            self.mutator.evaluate_best_individual(
                individual)  # TODO: all evaluating should be derived from Evaluator class

        self.selector.fast_nondominated_sorting(population.individuals)
        self.problem.populations.append(population)

        t_s = time.time()
        self.problem.logger.info("PSO: {}/{}".format(self.options['max_population_number'],
                                                     self.options['max_population_size']))

        i = 0
        while i < self.options['max_population_number']:
            offsprings = self.selector.select(population.individuals)

            pareto_front = []
            for individual in offsprings:
                if individual.features['front_number'] == 1:
                    pareto_front.append(individual)

            for individual in offsprings:
                index = randint(0, len(pareto_front) - 1)  # takes random individual from Pareto front
                best_individual = pareto_front[index]
                if best_individual is not individual:
                    self.mutator.update(best_individual)
                    self.mutator.mutate(individual)

            self.evaluate(offsprings)
            self.add_features(offsprings)

            for individual in offsprings:
                self.mutator.evaluate_best_individual(individual)

            self.selector.fast_nondominated_sorting(offsprings)
            population = Population(offsprings)
            self.problem.populations.append(population)

            i += 1

        t = time.time() - t_s
        self.problem.logger.info("PSO: elapsed time: {} s".format(t))



import platform
from abc import *
from enum import Enum
from random import uniform


class Individual(metaclass=ABCMeta):
    """ Collects information about one point in design space. """

    class State(Enum):
        EMPTY = 0
        IN_PROGRESS = 1
        EVALUATED = 2
        FAILED = 3

        @staticmethod
        def to_string(state):
            if state == Individual.State.EMPTY:
                return "Individual is empty."
            elif state == Individual.State.IN_PROGRESS:
                return "Individual is in progress."
            elif state == Individual.State.EVALUATED:
                return "Individual is evaluated."
            elif state == Individual.State.FAILED:
                return "Individual evaluation failed."

    def __init__(self, vector: list):
        self.vector = vector.copy()
        self.costs = []
        self.costs_signed = []
        self.state = self.State.EMPTY
        self.parents = []
        self.children = []
        self.features = {}
        self.custom = {}
        self.feasible = 0.0  # the distance from the feasibility region in min norm, its an index, not a
        self.precision = 7   # the default value of the considered decimals

        self.info = {"start_time": 0.0,
                     "finish_time": 0.0,
                     "population": -1,
                     "processor": platform.processor(),
                     "system": platform.system(),
                     "python": platform.python_version(),
                     "hostname": platform.node()}

    def calc_signed_costs(self, p_signs):
        """
        This function calculates the signed costs for every vector and insert the feasibility after
        :return:
        """
        self.costs_signed = list(map(lambda x, y: x * round(y, ndigits=self.precision), p_signs, self.costs))
        self.costs_signed.append(self.feasible)

    def __repr__(self):
        """ :return: [vector[p1, p2, ... pn]; costs[c1, c2, ... cn]] """

        string = "vector: ["
        for i, number in enumerate(self.vector):
            string += str(number)
            if i < len(self.vector) - 1:
                string += ", "
        string = string[:len(string) - 1]
        string += "]"

        string += "; costs:["
        for i, number in enumerate(self.costs):
            string += str(number)
            if i < len(self.costs) - 1:
                string += ", "
        string += "]"

        if len(self.custom) > 0:
            string += "; custom:["
            for i, number in enumerate(self.custom):
                string += str(number)
                if i < len(self.custom) - 1:
                    string += ", "
            string += "]"

        string += "; state: '{}', ".format(Individual.State.to_string(self.state))
        # string += "; info: {}, ".format(self.info)

        return string

    def __eq__(self, other):

        if self.costs_signed == []:
            diff = set(self.vector) - set(other.vector)
            return diff == set()
        else:
            diff = set(self.costs_signed) - set(other.costs_signed)
            return  diff == set()

    def __hash__(self):
        #return id(self)
        return hash(tuple(self.vector))

    def sync(self, individual):
        self.vector = individual.vector
        self.costs = individual.costs
        self.costs_signed = individual.costs_signed
        self.custom = individual.custom
        self.state = individual.state
        self.info = individual.info
